"""Read pv generation functions."""

import logging
import uuid
from datetime import datetime
from typing import List, Optional, Union

from sqlalchemy import func
from sqlalchemy.orm import Session, contains_eager

from pvsite_datamodel.pydantic_models import GenerationSum
from pvsite_datamodel.sqlmodels import (
    GenerationSQL,
    LocationGroupLocationSQL,
    LocationGroupSQL,
    LocationSQL,
    UserSQL,
)

logger = logging.getLogger(__name__)


def get_pv_generation_by_user_uuids(
    session: Session,
    start_utc: Optional[datetime] = None,
    end_utc: Optional[datetime] = None,
    user_uuids: Optional[List[str]] = None,
) -> List[GenerationSQL]:
    """Get the generation data by user uuids.

    :param session: database session
    :param start_utc: search filters >= on 'datetime_utc'. Can be None
    :param end_utc: search filters < on 'datetime_utc'. Can be None
    :param user_uuids: optional list of user uuids
    :return:list of pv yields.
    """
    # start main query
    query = session.query(GenerationSQL)
    query = query.join(LocationSQL)

    # Filter by time interval
    if start_utc is not None:
        query = query.filter(
            GenerationSQL.start_utc >= start_utc,
        )

    if end_utc is not None:
        query = query.filter(
            GenerationSQL.end_utc < end_utc,
        )

    if user_uuids is not None:
        query = query.join(LocationGroupLocationSQL)
        query = query.join(LocationGroupSQL)
        query = query.join(UserSQL)
        query = query.filter(UserSQL.user_uuid.in_(user_uuids))

    query = query.order_by(
        LocationSQL.location_uuid,
        GenerationSQL.start_utc,
    )

    # get all results
    generations: List[GenerationSQL] = query.all()

    return generations


def get_pv_generation_by_sites(
    session: Session,
    start_utc: Optional[datetime] = None,
    end_utc: Optional[datetime] = None,
    site_uuids: Optional[List[uuid.UUID]] = None,
    sum_by: Optional[str] = None,
) -> Union[List[GenerationSQL], List[GenerationSum]]:
    """Get the generation data by site.

    :param session: database session
    :param start_utc: search filters >= on 'datetime_utc'
    :param end_utc: search fileters < on 'datetime_utc'
    :param site_uuids: optional list of site uuids
    :param sum_by: optional string to sum by. Must be one of ['total', 'dno', 'gsp']
    :return: list of pv yields
    """

    if sum_by not in ["total", "dno", "gsp", None]:
        raise ValueError(f"sum_by must be one of ['total', 'dno', 'gsp'], not {sum_by}")

    query = session.query(GenerationSQL)
    query = query.join(LocationSQL)

    # Filter by time interval
    if start_utc is not None:
        query = query.filter(
            GenerationSQL.start_utc >= start_utc,
        )

    if end_utc is not None:
        query = query.filter(
            GenerationSQL.end_utc < end_utc,
        )

    if site_uuids is not None:
        query = query.filter(LocationSQL.location_uuid.in_(site_uuids))

    query = query.order_by(LocationSQL.location_uuid, GenerationSQL.start_utc)

    # make sure this is all loaded
    query = query.options(contains_eager(GenerationSQL.location)).populate_existing()

    if sum_by is None:
        # get all results
        generations: List[GenerationSQL] = query.all()
    else:
        subquery = query.subquery()

        group_by_variables = [subquery.c.start_utc]
        if sum_by == "dno":
            group_by_variables.append(LocationSQL.dno)
        if sum_by == "gsp":
            group_by_variables.append(LocationSQL.gsp)
        query_variables = group_by_variables.copy()
        query_variables.append(func.sum(subquery.c.generation_power_kw))

        query = session.query(*query_variables)
        query = query.join(LocationSQL)
        query = query.group_by(*group_by_variables)
        query = query.order_by(*group_by_variables)
        generations_raw = query.all()

        generations: List[GenerationSum] = []
        for generation_raw in generations_raw:
            if len(generation_raw) == 2:
                generation = GenerationSum(
                    start_utc=generation_raw[0], power_kw=generation_raw[1], name="total"
                )
            else:
                generation = GenerationSum(
                    start_utc=generation_raw[0], power_kw=generation_raw[2], name=generation_raw[1]
                )
            generations.append(generation)

    return generations



"""
Write helpers for the Generation table.
"""

import datetime as dt
import logging

import pandas as pd
from sqlalchemy.orm import Session

from pvsite_datamodel.sqlmodels import GenerationSQL
from pvsite_datamodel.write.utils import _insert_do_nothing_on_conflict

_log = logging.getLogger(__name__)


def insert_generation_values(
    session: Session,
    df: pd.DataFrame,
):
    """Insert a dataframe of generation values into the database.

    :param session: sqlalchemy session for interacting with the database
    :param df: dataframe with the data to insert
    """

    # rename site_uuid to location_uuid
    if "site_uuid" in df.columns and "location_uuid" not in df.columns:
        df.rename(columns={"site_uuid": "location_uuid"}, inplace=True)

    # Check for duplicated (site_uuid, start_utc) entries.
    # TODO Should we have a unique constraint on those in the Database instead?
    duplicates = df[df.duplicated(["location_uuid", "start_utc"])]
    sites_with_duplicate_times = duplicates["location_uuid"].unique()
    if len(sites_with_duplicate_times) > 0:
        for site_uuid in sites_with_duplicate_times:
            _log.warning(f'duplicate target datetimes for site "{site_uuid}"')

    # Build a list of the
    generation_sqls: list[dict] = []

    for _, row in df.iterrows():
        location_uuid = row["location_uuid"]
        start_utc = row["start_utc"]
        power_kw = row["power_kw"]
        end_utc = row.get("end_utc", default=start_utc + dt.timedelta(minutes=5))

        # Create a GenerationSQL object for each generation, and surface as dict
        generation = GenerationSQL(
            location_uuid=location_uuid,
            generation_power_kw=power_kw,
            start_utc=start_utc,
            # TODO This is arbitrary and should be fixed
            # https://github.com/openclimatefix/pv-site-datamodel/issues/52
            end_utc=end_utc,
        ).__dict__

        generation_sqls.append(generation)

    _insert_do_nothing_on_conflict(session, GenerationSQL, generation_sqls)



"""
Useful functions for write operations.
"""

from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import Session

from pvsite_datamodel.sqlmodels import Base


def _insert_do_nothing_on_conflict(session: Session, table: Base, rows: list[dict]):
    """Upserts rows into table.

    This functions checks the primary keys and constraints, and if present, does nothing
    :param session: sqlalchemy Session
    :param table: the table
    :param rows: the rows we are going to update
    """
    stmt = postgresql.insert(table.__table__)
    stmt = stmt.on_conflict_do_nothing()
    session.execute(stmt, rows)



""" Tools for making fake users and sites in the database."""

import json
import logging
from datetime import datetime, timezone
from typing import Optional, Tuple
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy import text
from sqlalchemy.orm.session import Session
from sqlalchemy.sql.functions import func

from pvsite_datamodel.pydantic_models import PVSiteEditMetadata
from pvsite_datamodel.read import get_or_create_model, get_site_by_uuid, get_user_by_email
from pvsite_datamodel.sqlmodels import (
    ForecastSQL,
    ForecastValueSQL,
    LocationAssetType,
    LocationGroupSQL,
    LocationSQL,
    UserSQL,
)
from pvsite_datamodel.write.data.dno import get_dno
from pvsite_datamodel.write.data.gsp import get_gsp

logger = logging.getLogger(__name__)


def make_fake_site(db_session, ml_id=1):
    """Make a site.

    This is mainly used for testing purposes.
    """

    site = LocationSQL(
        client_location_id=1,
        latitude=51,
        longitude=3,
        capacity_kw=4,
        inverter_capacity_kw=4,
        module_capacity_kw=4.3,
        created_utc=datetime.now(timezone.utc),
        ml_id=ml_id,
    )
    db_session.add(site)
    db_session.commit()

    return site


def create_site_group(db_session, site_group_name="test_site_group"):
    """Make a site group.

    This is mainly used for testing purposes.
    """
    # create site group
    site_group = LocationGroupSQL(location_group_name=site_group_name)
    db_session.add(site_group)
    db_session.commit()

    return site_group


# make site
def create_site(
    session: Session,
    client_site_id: int,
    client_site_name: str,
    latitude: float,
    longitude: float,
    capacity_kw: float,
    dno: Optional[str] = None,
    gsp: Optional[str] = None,
    country: Optional[str] = "uk",
    region: Optional[str] = None,
    asset_type: Optional[str] = LocationAssetType.pv.name,
    orientation: Optional[float] = None,
    tilt: Optional[float] = None,
    inverter_capacity_kw: Optional[float] = None,
    module_capacity_kw: Optional[float] = None,
    client_uuid: Optional[UUID] = None,
    ml_id: Optional[int] = None,
    user_uuid: Optional[str] = None,
) -> [LocationSQL, str]:
    """
    Create a site and adds it to the database.

    :param session: database session
    :param client_site_id: id the client uses for the site
    :param client_site_name: name the client uses for the site
    :param latitude: latitude of site as an integer
    :param longitude: longitude of site as an integer
    :param capacity_kw: capacity of site in kw
    :param dno: dno of site
    :param gsp: gsp of site
    :param country: country of site, default is uk
    :param region: region of site
    :param asset_type: type of asset (accepts "pv" or "wind")
    :param orientation: orientation of site, default is 180
    :param tilt: tilt of site, default is 35
    :param inverter_capacity_kw: inverter capacity of site in kw
    :param module_capacity_kw: module capacity of site in kw
    :param ml_id: internal ML modelling id
    :param user_uuid: the UUID of the user creating the site

    """
    set_session_user(session, user_uuid)

    max_ml_id = session.query(func.max(LocationSQL.ml_id)).scalar()

    if max_ml_id is None:
        max_ml_id = 0

    if country in [None, ""]:
        country = "uk"

    # if region in [None, ""]:
    #     region = "uk"

    if asset_type not in LocationAssetType.__members__:
        raise ValueError(
            f"""Invalid asset_type. Received: {asset_type},
            but must one of ({', '.join(map(lambda type: type.name, LocationAssetType))})"""
        )

    if orientation in [None, ""]:
        orientation = 180

    if tilt in [None, ""]:
        tilt = 35

    if inverter_capacity_kw in [None, ""]:
        inverter_capacity_kw = capacity_kw

    if module_capacity_kw in [None, ""]:
        module_capacity_kw = capacity_kw

    if gsp is None:
        gsp = get_gsp(latitude=latitude, longitude=longitude)
        gsp = json.dumps(gsp)

    if dno is None:
        dno = get_dno(latitude=latitude, longitude=longitude)
        dno = json.dumps(dno)

    site = LocationSQL(
        ml_id=ml_id if ml_id else max_ml_id + 1,
        client_location_id=client_site_id,
        client_location_name=client_site_name,
        latitude=latitude,
        longitude=longitude,
        capacity_kw=capacity_kw,
        dno=dno,
        gsp=gsp,
        country=country,
        region=region,
        asset_type=asset_type,
        orientation=orientation,
        tilt=tilt,
        inverter_capacity_kw=inverter_capacity_kw,
        module_capacity_kw=module_capacity_kw,
        client_uuid=client_uuid,
    )

    session.add(site)

    session.commit()

    message = (
        f"Site with client location id {site.client_location_id} "
        f"and location uuid {site.location_uuid} created successfully"
    )

    return site, message


def create_user(
    session: Session,
    email: str,
    site_group_name: str,
) -> UserSQL:
    """Create a user.

    :param session: database session
    :param email: email of user being created
    :param site_group_name: name of the site group this user will be part of
    """

    site_group = (
        session.query(LocationGroupSQL)
        .filter(LocationGroupSQL.location_group_name == site_group_name)
        .first()
    )

    user = UserSQL(email=email, location_group_uuid=site_group.location_group_uuid)

    session.add(user)
    site_group.users.append(user)
    session.commit()

    return user


# update functions for site and site group
def add_site_to_site_group(
    session: Session, site_uuid: str, site_group_name: str
) -> LocationGroupSQL:
    """Add a site to a site group.

    NB: Sites can belong to many site groups.
    :param session: database session
    :param site_uuid: uuid of site
    :param site_group_name: name of site group
    """
    site_group = (
        session.query(LocationGroupSQL)
        .filter(LocationGroupSQL.location_group_name == site_group_name)
        .first()
    )

    site = session.query(LocationSQL).filter(LocationSQL.location_uuid == site_uuid).one()

    if site not in site_group.locations:
        site_group.locations.append(site)

    session.commit()

    return site_group.locations


def remove_site_from_site_group(
    session: Session, site_uuid: str, site_group_name: str
) -> [LocationSQL]:
    """Remove a site to a site group.

    NB: Sites can belong to many site groups.
    :param session: database session
    :param site_uuid: uuid of site
    :param site_group_name: name of site group
    """
    site_group = (
        session.query(LocationGroupSQL)
        .filter(LocationGroupSQL.location_group_name == site_group_name)
        .first()
    )

    site = session.query(LocationSQL).filter(LocationSQL.location_uuid == site_uuid).one()

    if site in site_group.locations:
        new_sites = [site for site in site_group.locations if str(site.location_uuid) != site_uuid]
        site_group.locations = new_sites

    session.commit()

    return site_group.locations


# change site group for user
def change_user_site_group(session, email: str, site_group_name: str):
    """
    Change user to a specific site group name.

    :param session: the database session
    :param email: the email of the user
    :param site_group_name: the name of the site group
    """
    update_user_site_group(session=session, email=email, site_group_name=site_group_name)
    user = get_user_by_email(session=session, email=email)
    user_site_group = user.location_group.location_group_name
    user = user.email
    return user, user_site_group


def update_user_site_group(session: Session, email: str, site_group_name: str) -> UserSQL:
    """Change site group for user.

    :param session: database session
    :param email: email of user
    :param site_group_name: name of site group
    """
    site_group = (
        session.query(LocationGroupSQL)
        .filter(LocationGroupSQL.location_group_name == site_group_name)
        .first()
    )

    user = session.query(UserSQL).filter(UserSQL.email == email)

    user = user.update({"location_group_uuid": site_group.location_group_uuid})

    session.commit()

    return user


# update site metadata
def edit_site(
    session: Session, site_uuid: str, site_info: PVSiteEditMetadata, user_uuid: str = None
) -> Tuple[LocationSQL, str]:
    """
    Edit an existing site. Fill in only the fields that need to be updated.

    :param session: database session
    :param site_uuid: the existing site uuid
    :param site_info: metadata for editing the site, should be an instance of PVSiteEditMetadata
        - client_site_id: id the client uses for the site
        - client_site_name: name the client uses for the site
        - orientation: orientation of site
        - tilt: tilt of site
        - latitude: latitude of site as an integer
        - longitude: longitude of site as an integer
        - inverter_capacity_kw: inverter capacity of site in kw
        - module_capacity_kw: module capacity of site in kw
        - capacity_kw: capacity of site in kw
        - dno: dno of site
        - gsp: gsp of site
        - client_uuid: The UUID of the client this site belongs to
    :param user_uuid: the UUID of the user editing the site
    """
    set_session_user(session, user_uuid)

    site = session.query(LocationSQL).filter(LocationSQL.location_uuid == site_uuid).first()

    update_data = site_info.model_dump(exclude_unset=True, exclude_none=False)

    if "client_site_name" in update_data:
        site.client_location_name = update_data.pop("client_site_name")
    if "client_site_id" in update_data:
        site.client_location_id = update_data.pop("client_site_id")

    # Update model class variable from requested fields
    for var, value in update_data.items():
        setattr(site, var, value)

    session.add(site)
    session.commit()
    session.refresh(site)

    message = f"Location with location uuid {site.location_uuid} updated successfully"

    return site, message


def set_site_to_inactive_if_not_in_site_group(
    session: Session, site_uuid: str, user_uuid: str, ignore_ocf_site_group: Optional[bool] = True
):
    """Set to inactive, if not in a site group.

    :param session: database session
    :param site_uuid: the site uuid
    :param user_uuid: the UUID of the user setting the site to inactive
    :param ignore_ocf_site_group: if True,
        ignore the "ocf" site group when checking if the site is in a site group
    """

    set_session_user(session, user_uuid)

    # get site
    site = session.query(LocationSQL).filter(LocationSQL.location_uuid == site_uuid).first()

    # check if site is in any other site group
    site_groups: [LocationGroupSQL] = site.location_groups
    if ignore_ocf_site_group:
        # only look at sites groups that are not called "ocf"
        site_groups = [sg for sg in site_groups if sg.location_group_name != "ocf"]

    logger.info(f"Site {site_uuid} is in {len(site_groups)} site groups.")

    if len(site_groups) == 0:
        logger.info(f"Site {site_uuid} is not in any site group, setting it to inactive.")

        site.active = False
        session.commit()


# delete functions for site, user, and site group
def delete_site(
    session: Session, site_uuid: str, user_uuid: Optional[str] = None
) -> LocationGroupSQL:
    """Delete a site group.

    :param session: database session
    :param site_uuid: unique identifier for site
    :param user_uuid: the UUID of the user deleting the site
    """
    set_session_user(session, user_uuid)

    site = session.query(LocationSQL).filter(LocationSQL.location_uuid == site_uuid).first()

    forecast_uuids = (
        session.query(ForecastSQL.forecast_uuid)
        .filter(ForecastSQL.location_uuid == site_uuid)
        .all()
    )

    forecast_uuids = [str(forecast_uuid[0]) for forecast_uuid in forecast_uuids]

    # get and delete all forecast values for site
    stmt = sa.delete(ForecastValueSQL).where(ForecastValueSQL.forecast_uuid.in_(forecast_uuids))
    session.execute(stmt)

    stmt_2 = sa.delete(ForecastSQL).where(ForecastSQL.forecast_uuid.in_(forecast_uuids))
    session.execute(stmt_2)

    # we decided not to delete generation data for the site
    # because it seems sensible to keep it for now

    session.delete(site)

    message = f"Location with location uuid {site.location_uuid} deleted successfully"

    session.commit()

    return message


# delete user
def delete_user(session: Session, email: str) -> str:
    """Delete a user.

    :param session: database session
    :param email: email of user being deleted
    """
    user = session.query(UserSQL).filter(UserSQL.email == email).first()

    session.delete(user)

    message = (
        f"User with email {user.email} and location_group_uuid "
        f"{user.location_group_uuid} deleted successfully"
    )

    session.commit()

    return message


# delete site group
def delete_site_group(session: Session, site_group_name: str) -> str:
    """Delete a site group.

    :param session: database session
    :param site_group_name: name of site group being deleted
    """
    site_group = (
        session.query(LocationGroupSQL)
        .filter(LocationGroupSQL.location_group_name == site_group_name)
        .first()
    )

    site_group_users = site_group.users

    if len(site_group_users) > 0:
        message = (
            f"Location group with name {site_group.location_group_name} and "
            f"location group uuid {site_group.location_group_uuid} has users and cannot be deleted."
        )
        return message

    session.delete(site_group)

    message = (
        f"Location group with name {site_group.location_group_name} "
        f"and location group uuid {site_group.location_group_uuid} deleted successfully."
    )

    session.commit()

    return message


def assign_model_name_to_site(session: Session, site_uuid, model_name):
    """Assign model to site.

    :param session: database session
    :param site_uuid: site uuid
    :param model_name: name of the model
    """

    site = get_site_by_uuid(session=session, site_uuid=site_uuid)

    model = get_or_create_model(session=session, name=model_name)

    site.ml_model_uuid = model.model_uuid
    session.commit()


def set_session_user(session: Session, user_uuid: str):
    """Set user session variable.

    Sets a variable which is then used by the log_site_changes function when updating
    the site history table.

    :param session: the session
    :param user_uuid: the user UUID
    """
    if user_uuid is not None:
        session.execute(text(f"SET pvsite_datamodel.current_user_uuid = '{user_uuid}'"))
    else:
        session.execute(text("RESET pvsite_datamodel.current_user_uuid"))


def add_child_location_to_parent_location(
    session: Session, child_location_uuid: str, parent_location_uuid: str
) -> None:
    """Add a child location to a parent location.

    :param session: database session
    :param child_location_uuid: the UUID of the child location
    :param parent_location_uuid: the UUID of the parent location
    """
    child_location = get_site_by_uuid(session=session, site_uuid=child_location_uuid)
    parent_location = get_site_by_uuid(session=session, site_uuid=parent_location_uuid)

    if parent_location not in child_location.parent_locations:
        parent_location.child_locations.append(child_location)
        # note this also adds the parent to the child_location.parent_locations

    session.commit()



"""SQLAlchemy definition of the pvsite database schema."""

from __future__ import annotations

# This means we can use Typing of objects that have jet to be defined
import enum
import uuid
from datetime import datetime
from typing import List, Optional

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB, UUID
from sqlalchemy.orm import Mapped, declarative_base, relationship
from sqlalchemy.schema import UniqueConstraint

Base = declarative_base()


class CreatedMixin:
    """Mixin to add created datetime to model."""

    created_utc = sa.Column(sa.DateTime, default=lambda: datetime.utcnow())


class MLModelSQL(Base, CreatedMixin):
    """ML model that is being used."""

    __tablename__ = "ml_model"

    model_uuid = sa.Column(UUID, primary_key=True, server_default=sa.func.gen_random_uuid())
    name = sa.Column(sa.String)
    version = sa.Column(sa.String)

    # Add description with default message
    description = sa.Column(sa.String, nullable=True, server_default="No description provided.")

    forecast_values: Mapped[List["ForecastValueSQL"]] = relationship(
        "ForecastValueSQL", back_populates="ml_model"
    )

    locations: Mapped[List["LocationSQL"]] = relationship("LocationSQL", back_populates="ml_model")


class UserSQL(Base, CreatedMixin):
    """Class representing the users table.

    Each user row specifies a single user.
    """

    __tablename__ = "users"
    __tables_args__ = (UniqueConstraint("email", name="idx_email"),)

    user_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    email = sa.Column(sa.String(255), index=True, unique=True)
    location_group_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("location_groups.location_group_uuid"),
        nullable=False,
        comment="The foreign key to the location_groups table",
    )

    # Relationships
    location_group: Mapped["LocationGroupSQL"] = relationship(
        "LocationGroupSQL", back_populates="users"
    )
    api_request = relationship("APIRequestSQL", back_populates="user")


class LocationGroupSQL(Base, CreatedMixin):
    """Class representing the location_groups table.

    Each location_group row specifies a single group of locations.
    """

    __tablename__ = "location_groups"

    location_group_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    location_group_name = sa.Column(sa.String(255), index=True, unique=True)
    service_level = sa.Column(
        sa.Integer,
        default=0,
        comment="The service level of the location group. 0 is free, 1 is paid.",
    )

    # Relationships
    # N-N
    locations: Mapped[List["LocationSQL"]] = relationship(
        "LocationSQL", secondary="location_group_locations", back_populates="location_groups"
    )
    # 1-N, one location group can have many users
    users: Mapped[List[UserSQL]] = relationship("UserSQL", back_populates="location_group")


class LocationGroupLocationSQL(Base, CreatedMixin):
    """Class representing the location_group_locations table.

    Each location_group_location row specifies a single location in a location group.
    """

    __tablename__ = "location_group_locations"
    __table_args__ = (
        UniqueConstraint(
            "location_group_uuid", "location_uuid", name="idx_location_group_location"
        ),
    )

    location_group_location_uuid = sa.Column(
        UUID(as_uuid=True), default=uuid.uuid4, primary_key=True
    )
    location_group_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("location_groups.location_group_uuid"),
        nullable=False,
        comment="The foreign key to the location_groups table",
    )
    location_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("locations.location_uuid"),
        nullable=False,
        comment="The foreign key to the locations table",
    )


class LocationLocationSQL(Base, CreatedMixin):
    """Class representing the location_locations table.

    Each location_to_location row specifies a single location in a location.
    We want to to be able to attaches multiple sites to several regions.
    E.g. pv solar sites in a GSP region, and a National Grid region.
    """

    __tablename__ = "location_locations"
    __table_args__ = (
        UniqueConstraint(
            "location_child_uuid", "location_parent_uuid", name="idx_location_location"
        ),
    )

    location_location_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    location_parent_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("locations.location_uuid"),
        nullable=False,
        comment="The foreign key to the locations table",
    )
    location_child_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("locations.location_uuid"),
        nullable=False,
        comment="The foreign key to the locations table",
    )


class LocationAssetType(enum.Enum):
    """Enum type representing a location's asset type."""

    pv = 1
    wind = 2


class LocationType(enum.Enum):
    """Enum type representing a location's location type."""

    site = 1
    region = 2


class LocationSQL(Base, CreatedMixin):
    """Class representing the locations table.

    Each location row specifies location in space. It can represent
    - a solar single panel
    - cluster of solar panels found on a residential house or commercial building.
    - a large solar farm
    - a wind farm

    # TODO update this to LocationSQL in future release + other renames
    """

    __tablename__ = "locations"

    location_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    client_location_id = sa.Column(
        sa.Integer, index=True, comment="The ID of the location as given by the providing client"
    )
    client_location_name = sa.Column(
        sa.String(255),
        index=True,
        comment="The ID of the location as given by the providing client",
    )

    country = sa.Column(
        sa.String(255), server_default="uk", comment="The country in which the location is located"
    )

    asset_type = sa.Column(
        sa.Enum(LocationAssetType, name="site_asset_type"),
        nullable=False,
        server_default=LocationAssetType.pv.name,
    )

    latitude = sa.Column(sa.Float)
    longitude = sa.Column(sa.Float)
    capacity_kw = sa.Column(
        sa.Float, comment="The physical limit on the production capacity of the location"
    )

    ml_id = sa.Column(
        sa.Integer,
        autoincrement=True,
        nullable=False,
        comment="Auto-incrementing integer ID of the location for use in ML training",
    )

    active = sa.Column(
        sa.Boolean,
        default=True,
        unique=False,
        comment="Indicates if location is active",
    )

    client_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("clients.client_uuid"),
        nullable=True,
        index=True,
        comment="The UUID of the client this location belongs to",
    )

    ml_model_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("ml_model.model_uuid"),
        nullable=True,
        comment="The ML Model which should be used for this location",
    )
    # new fields
    location_type = sa.Column(
        sa.Enum(LocationType, name="location_type"),
        nullable=False,
        server_default=LocationType.site.name,
    )
    location_metadata = sa.Column(
        JSONB,
        nullable=True,
        comment="Specific properties of the location, "
        "for example for a location, the tilt and orientation of the "
        "solar panels, or for a region, the region name.",
    )

    # legacy fields
    region = sa.Column(
        sa.String(255), comment="The region within the country in which the location is located"
    )
    dno = sa.Column(sa.String(255), comment="The Distribution Node Operator that owns the location")
    gsp = sa.Column(
        sa.String(255), comment="The Grid Supply Point in which the location is located"
    )
    # For metadata `NULL` means "we don't know".
    orientation = sa.Column(
        sa.Float, comment="The rotation of the panel in degrees. 180° points south"
    )
    tilt = sa.Column(
        sa.Float, comment="The tile of the panel in degrees. 90° indicates the panel is vertical"
    )
    inverter_capacity_kw = sa.Column(sa.Float, comment="The inverter capacity of the location")
    module_capacity_kw = sa.Column(
        sa.Float, comment="The PV module nameplate capacity of the location"
    )

    # relationships
    forecasts: Mapped[List["ForecastSQL"]] = relationship("ForecastSQL", back_populates="location")
    generation: Mapped[List["GenerationSQL"]] = relationship("GenerationSQL")
    inverters: Mapped[List["InverterSQL"]] = relationship(
        "InverterSQL", back_populates="location", cascade="all, delete-orphan"
    )
    location_groups: Mapped[List["LocationGroupSQL"]] = relationship(
        "LocationGroupSQL", secondary="location_group_locations", back_populates="locations"
    )
    client: Mapped[List["ClientSQL"]] = relationship("ClientSQL", back_populates="locations")
    ml_model: Mapped[Optional[MLModelSQL]] = relationship("MLModelSQL", back_populates="locations")

    # n:n mapping to reference back to locations.
    # This means many site can be part of a many different regions.
    child_locations: Mapped[List["LocationSQL"]] = relationship(
        "LocationSQL",
        secondary="location_locations",
        primaryjoin="LocationSQL.location_uuid == LocationLocationSQL.location_child_uuid",
        secondaryjoin="LocationSQL.location_uuid == LocationLocationSQL.location_parent_uuid",
        backref="parent_locations",
    )


class LocationHistorySQL(Base, CreatedMixin):
    """Class representing the locations history table.

    Stores a history of changes to locatios over time. Uses JSONB so that schema changes to the
    location table do not affect the history table.
    """

    __tablename__ = "locations_history"

    location_history_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)

    location_uuid = sa.Column(
        UUID(as_uuid=True),
        nullable=False,
        index=True,
        comment="The location which this history record relates to",
    )

    # JSONB column to store the snapshot of the location data
    location_data = sa.Column(
        JSONB, nullable=False, comment="A snapshot of the location record as JSONB"
    )

    # Foreign key to track the user who made the change
    changed_by = sa.Column(UUID(as_uuid=True), sa.ForeignKey("users.user_uuid"), nullable=True)

    operation_type = sa.Column(sa.TEXT, nullable=False)


class ClientSQL(Base, CreatedMixin):
    """Class representing the client table.

    Each client row specifies a single client.
    """

    __tablename__ = "clients"

    client_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    client_name = sa.Column(sa.String(255), nullable=False, index=True, unique=True)

    locations: Mapped[List[LocationSQL]] = relationship("LocationSQL", back_populates="client")


class GenerationSQL(Base, CreatedMixin):
    """Class representing the generation table.

    Each generation row specifies a generated power output over a
    given time range for a location.

    *Approximate size: *
    Generation populated every 5 minutes per location * 4000 locations = ~1,125,000 rows per day
    """

    __tablename__ = "generation"
    __table_args__ = (
        UniqueConstraint(
            "location_uuid", "start_utc", "end_utc", name="uniq_cons_location_start_end"
        ),
    )

    generation_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    location_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("locations.location_uuid"),
        nullable=False,
        index=True,
        comment="The location for which this generation yield belongs to",
    )
    generation_power_kw = sa.Column(
        sa.Float,
        nullable=False,
        comment="The actual generated power in kW at this location for this datetime interval",
    )

    start_utc = sa.Column(
        sa.DateTime,
        nullable=False,
        index=True,
        comment="The start of the time interval over which this generated power value applies",
    )
    end_utc = sa.Column(
        sa.DateTime,
        nullable=False,
        comment="The end of the time interval over which this generated power value applies",
    )

    location: Mapped["LocationSQL"] = relationship("LocationSQL", back_populates="generation")


class ForecastSQL(Base, CreatedMixin):
    """Class representing the forecasts table.

    Each forecast row refers to a sequence of predicted solar generation values
    over a set of target times for one location.

    *Approximate size: *
    One forecast per location every 5 minutes = ~1,125,000 rows per day
    """

    __tablename__ = "forecasts"

    forecast_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    location_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("locations.location_uuid"),
        nullable=False,
        comment="The location for which the forecast sequence was generated",
    )

    # The timestamp at which we are making the forecast. This is often referred as "now" in the
    # modelling code.
    # Note that this could be very different from the `created_utc` time, for instance if we
    # run the model for a given "now" timestamp in the past.
    timestamp_utc = sa.Column(
        sa.DateTime,
        nullable=False,
        comment="The creation time of the forecast sequence",
    )

    forecast_version = sa.Column(
        sa.String(32),
        nullable=False,
        comment="The semantic version of the model used to generate the forecast",
    )

    # one (forecasts) to many (forecast_values)
    forecast_values: Mapped["ForecastValueSQL"] = relationship("ForecastValueSQL")
    location = relationship("LocationSQL", back_populates="forecasts")

    __table_args__ = (
        # With this index, we are assuming that it doesn't make sense to do a query solely on
        # `timestamp_utc`: we always also filter by location_uuid.
        sa.Index("ix_forecasts_location_uuid_timestamp_utc", "location_uuid", "timestamp_utc"),
    )


class ForecastValueSQL(Base, CreatedMixin):
    """Class representing the forecast_values table.

    Each forecast_value row is a prediction for the power output
    of a location over a target datetime interval. Many predictions
    are made for each location at each target interval.

    *Approximate size: *
    One forecast value every 5 minutes per location per forecast.
    Each forecast's prediction sequence covers 24 hours of target
    intervals = ~324,000,000 rows per day
    """

    __tablename__ = "forecast_values"

    forecast_value_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)

    start_utc = sa.Column(
        sa.DateTime,
        nullable=False,
        index=True,
        comment="The start of the time interval over which this predicted power value applies",
    )
    end_utc = sa.Column(
        sa.DateTime,
        nullable=False,
        comment="The end of the time interval over which this predicted power value applies",
    )
    forecast_power_kw = sa.Column(
        sa.Float,
        nullable=False,
        comment="The predicted power generation of this location for the given time interval",
    )

    # This is the different between `start_utc` and the `forecast`'s `timestamp_utc`, in minutes.
    # It's useful to have it in its own column to efficiently query forecasts for a given horizon.
    horizon_minutes = sa.Column(
        sa.Integer,
        nullable=False,
        server_default=sa.text("-1"),
        comment=(
            "The time difference between the creation time of the forecast value "
            "and the start of the time interval it applies for"
        ),
    )

    forecast_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("forecasts.forecast_uuid"),
        nullable=False,
        comment="The forecast sequence this forcast value belongs to",
    )
    ml_model_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("ml_model.model_uuid"),
        nullable=True,
        comment="The ML Model this forcast value belongs to",
    )

    probabilistic_values = sa.Column(
        JSONB,
        nullable=False,
        server_default=sa.text("'{}'"),  # Default to an empty JSON object
        comment="Probabilistic forecast values, like p10, p50, p90",
    )

    forecast: Mapped["ForecastSQL"] = relationship("ForecastSQL", back_populates="forecast_values")
    ml_model: Mapped[Optional[MLModelSQL]] = relationship(
        "MLModelSQL", back_populates="forecast_values"
    )

    __table_args__ = (
        # Here we assume that we always filter on `horizon_minutes` *for given forecasts*.
        sa.Index(
            "ix_forecast_values_forecast_uuid_horizon_minutes", "forecast_uuid", "horizon_minutes"
        ),
    )


class StatusSQL(Base, CreatedMixin):
    """Class representing the status table.

    Each status row defines a message reporting on the status of the
    services within the nowcasting domain

    *Approximate size: *
    ~1 row per day
    """

    __tablename__ = "status"

    status_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    status = sa.Column(sa.String(255))
    message = sa.Column(sa.String(255))


class InverterSQL(Base, CreatedMixin):
    """Class representing the inverters table.

    Each InverterSQL row represents an inverter tied to a LocationSQL row.

    *Approximate size: *
    4 clients * ~1000 locations each * ~1 inverter each = ~4000 rows
    """

    __tablename__ = "inverters"

    inverter_uuid = sa.Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True)
    location_uuid = sa.Column(
        UUID(as_uuid=True),
        sa.ForeignKey("locations.location_uuid"),
        nullable=False,
        index=True,
        comment="The UUID for the location that has this inverter",
    )

    location: Mapped["LocationSQL"] = relationship("LocationSQL", back_populates="inverters")


class APIRequestSQL(Base, CreatedMixin):
    """Information about what API route was called."""

    __tablename__ = "api_request"

    uuid = sa.Column(UUID, primary_key=True, server_default=sa.func.gen_random_uuid())
    url = sa.Column(sa.String)

    user_uuid = sa.Column(UUID, sa.ForeignKey("users.user_uuid"), index=True)
    user = relationship("UserSQL", back_populates="api_request")



"""
Write helpers for the Forecast and ForecastValues table.
"""

import logging
from typing import Optional

import pandas as pd
from sqlalchemy.orm import Session

from pvsite_datamodel.read.model import get_or_create_model
from pvsite_datamodel.sqlmodels import ForecastSQL, ForecastValueSQL

_log = logging.getLogger(__name__)


def insert_forecast_values(
    session: Session,
    forecast_meta: dict,
    forecast_values_df: pd.DataFrame,
    ml_model_name: Optional[str] = None,
    ml_model_version: Optional[str] = None,
):
    """Insert a dataframe of forecast values and forecast meta info into the database.

    :param session: sqlalchemy session for interacting with the database
    :param forecast_meta: Meta info about the forecast values
    :param forecast_values_df: dataframe with the data to insert
    :param ml_model_name: name of the ML model used to generate the forecast
    :param ml_model_version: version of the ML model used to generate the forecast
    """

    if "site_uuid" in forecast_meta and "location_uuid" not in forecast_meta:
        forecast_meta["location_uuid"] = forecast_meta["site_uuid"]
        forecast_meta.pop("site_uuid")

    forecast = ForecastSQL(**forecast_meta)
    session.add(forecast)

    # Flush to get the Forecast's primary key.
    session.flush()

    if (ml_model_name is not None) and (ml_model_version is not None):
        ml_model = get_or_create_model(session, ml_model_name, ml_model_version)
        ml_model_uuid = ml_model.model_uuid
    else:
        ml_model_uuid = None

    rows = forecast_values_df.to_dict("records")
    session.bulk_save_objects(
        [
            ForecastValueSQL(
                **row,
                forecast_uuid=forecast.forecast_uuid,
                ml_model_uuid=ml_model_uuid,
            )
            for row in rows
        ]
    )
    session.commit()



"""
Tools for making clients in the database.
"""

import logging
from uuid import UUID

from sqlalchemy.orm.session import Session

from pvsite_datamodel.sqlmodels import ClientSQL, LocationSQL

_log = logging.getLogger(__name__)


def create_client(session: Session, client_name: str) -> ClientSQL:
    """Create a client.

    :param session: database session
    :param client_name: name of client being created
    """
    client = ClientSQL(client_name=client_name)

    session.add(client)
    session.commit()

    return client


def edit_client(session: Session, client_uuid: UUID, client_name: str) -> ClientSQL:
    """Edit an existing client.

    :param session: database session
    :param client_uuid: the existing client uuid
    :param client_name: name of the client
    """
    client = session.query(ClientSQL).filter(ClientSQL.client_uuid == client_uuid).first()

    client.client_name = client_name

    session.add(client)
    session.commit()

    return client


def assign_site_to_client(session: Session, site_uuid: str, client_name: str) -> str:
    """Assign site to client.

    :param session: database session
    :param site_uuid: uuid of site
    :param client_name: name of the client the site will be assigned to.
    """

    client = session.query(ClientSQL).filter(ClientSQL.client_name == client_name).first()

    site = session.query(LocationSQL).filter(LocationSQL.location_uuid == site_uuid).first()

    site.client_uuid = client.client_uuid

    session.add(site)
    session.commit()

    message = (
        f"Site with location uuid {site_uuid} successfully assigned to the client {client_name}"
    )

    return message
